// BFont.cpp (this is -*- C++ -*-)
// 
// \author: Bjoern Giesler <bjoern@giesler.de>
// 
// 
// 
// $Author: giesler $
// $Locker$
// $Revision$
// $Date: 2002-08-19 10:41:28 +0200 (Mon, 19 Aug 2002) $

/* system includes */
#include <nds.h>

/* my includes */
#include "BFont.h"
#include "BGraphics.h"

BFont::BFont(const BImage& img, const GlyphDescription indices[],
	     int firstchar, int lastchar):
  img(img), indices(indices), firstchar(firstchar), lastchar(lastchar)
{
}

int
BFont::widthOfGlyph(int glyph)
{
  if(glyph == ' ') return widthOfGlyph('f');
  else if(glyph < firstchar || glyph > lastchar) return -1;

  return indices[glyph-firstchar].width;
}

int
BFont::widthOfString(const std::string& str)
{
  int width = 0;
  for(unsigned int i=0; i<str.size(); i++)
    {
      int w = widthOfGlyph(str[i]);
      if(w > -1) width += (w+1);
    }

  return width;
}

BPoint
BFont::drawGlyph(BImage& img, const BPoint& point, int glyph, uint16 color,
		 const BRect& clip, TextDirection dir)
{
  BPoint pt = point;
  
  if(glyph == ' ')
    {
      switch(dir)
	{
	case DIR_LEFTTORIGHT:
	  pt.x += widthOfGlyph('f');
	  break;
	case DIR_RIGHTTOLEFT:
	  pt.x -= widthOfGlyph('f');
	  break;
	case DIR_TOPTOBOTTOM:
	  pt.y += widthOfGlyph('f');
	  break;
	case DIR_BOTTOMTOTOP:
	  pt.y -= widthOfGlyph('f');
	  break;
	}
      return pt;
    }

  int x = point.x, y = point.y;

  if(glyph < firstchar || glyph > lastchar) return point;

  int xstart = indices[glyph-firstchar].start;
  int xwidth = indices[glyph-firstchar].width;
  int xend = xstart + xwidth;

  int ystart = 0, yend = this->img.height();

  const uint8* srcBytes = this->img.bytes();
  unsigned int srcw = this->img.width();
  uint16* tgtBytes = (uint16*)img.bytes();
  unsigned int tgtw = img.width(), tgth = img.height();

  for(int r=ystart; r<yend; r++)
    for(int c=xstart; c<xend; c++)
      {
	int tgtx=0, tgty=0;

	switch(dir)
	  {
	  case DIR_LEFTTORIGHT:
	    tgtx = x + (c - xstart);
	    tgty = y + (r - ystart);
	    break;
	  case DIR_TOPTOBOTTOM:
	    tgtx = x - (r - ystart);
	    tgty = y + (c - xstart);
	    break;
	  case DIR_RIGHTTOLEFT:
	    tgtx = x - (c - xstart);
	    tgty = y - (r - ystart);
	    break;
	  case DIR_BOTTOMTOTOP:
	    tgtx = x + (r - ystart);
	    tgty = y - (c - xstart);
	    break;
	  }

	if(tgtx < 0 || tgtx < clip.pt.x ||
	   tgtx >= (int)tgtw || tgtx >= (int)(clip.pt.x + clip.sz.width) ||
	   tgty < 0 || tgty < clip.pt.y ||
	   tgty >= (int)tgth || tgty >= (int)(clip.pt.y + clip.sz.height))
	  continue;

	uint8 srcbyte = srcBytes[r*srcw+c];
	if(srcbyte >= 0x1f) // maxval
	  continue;

	uint16& tgtbyte = tgtBytes[tgty*tgtw+tgtx];
	tgtbyte = alphablend(color, tgtbyte, 0x1f-srcbyte);
      }


  switch(dir)
    {
    case DIR_LEFTTORIGHT:
      pt.x += (xwidth + 1);
      break;
    case DIR_TOPTOBOTTOM:
      pt.y += (xwidth + 1);
      break;
    case DIR_RIGHTTOLEFT:
      pt.x -= (xwidth + 1);
      break;
    case DIR_BOTTOMTOTOP:
      pt.y -= (xwidth + 1);
    }

  return pt;
}

BPoint
BFont::drawString(BImage& img, const BPoint& point, const std::string& str,
		  uint16 color, const BRect& clip, TextDirection dir)
{
  BPoint pt = point;
  for(unsigned int i=0; i<str.size(); i++)
    pt = drawGlyph(img, pt, str[i], color, clip, dir);
  return pt;
}

BFont::GlyphDescription Sansserif_12_indices[] = {
  {2,1}, {6,3}, {11, 5}, {18,5}, {25,9}, {36,6}, {42,1},
  {46,3}, {49,3}, {54,5}, {61,5}, {67,2}, {71,3}, {76,1}, {78,4},

  // 0-9
  {83,5}, {91,3}, {97,5}, {104,5}, {111,5}, {118,5}, {125,5},
  {132,5}, {139,5}, {146,5},

  {153,1}, {156,2}, {161,5}, {168,5}, {175,5}, {182,4}, {188,8},

  // A-Z
  {197,8}, {206,5}, {213,6}, {221,6}, {229,5}, {236,5}, {242,6},
  {250,6}, {258,3}, {262,4}, {268,5}, {275,5}, {281,7}, {290,6},
  {298,7}, {307,5}, {314,7}, {323,6}, {330,5}, {336,7}, {344,6},
  {351,8}, {360,7}, {368,7}, {376,5}, {383,5},

  {390,2}, {393,4}, {398,2}, {402,5}, {408,6}, {416,2},

  // a-z
  {421,4}, {427,4}, {433,4}, {439,4}, {445,4}, {450,4}, {454,4},
  {460,4}, {466,1}, {467,3}, {472,4}, {478,1}, {481,7}, {490,4},
  {496,4}, {502,4}, {508,4}, {514,3}, {518,4}, {523,4}, {528,4},
  {533,6}, {540,5}, {546,6}, {552,6}, {559,3},

  {563,5}, {571,1}, {575,5}, {581,5}
};

int Sansserif_12_width = 587;
int Sansserif_12_height = 12;
int Sansserif_12_firstchar = '!';
int Sansserif_12_lastchar = '~';

#define EXTERNAL_DATA(name) \
  extern const uint8 name[]; \
  extern const uint8 name##_end[]; \
  extern const uint32 name##_size
EXTERNAL_DATA(Sansserif_12_Antialias_bin);

BFont*
BFont::ss12Font()
{
  if(!ss12)
    {
      ss12Img = new BImage(0, 0, Sansserif_12_width, Sansserif_12_height,
			   (uint8*)Sansserif_12_Antialias_bin,
			   BImage::FMT_GREY5, false);
      ss12 = new BFont(*ss12Img, Sansserif_12_indices,
		       Sansserif_12_firstchar, Sansserif_12_lastchar);
    }
  return ss12;
}

BFont* BFont::ss12 = NULL;
BImage* BFont::ss12Img = NULL;
